/****************************************************************
*   Utility routines for generic C                              *
*****************************************************************
*   Text Based Menu & Windowing                                 *
*                                                               *
*                                     Yudong Yang               *
*     Dept. of Computer Science and Technology, TsingHua Univ.  *
*   File : menu.c                                               *
****************************************************************/
//$$$ remarks begin with $$$ is system dependent codes
//$$$ for code portion to systems other than MSDOS and Watcom C386
//$$$ you must change those sections accordingly  

// include files
#include <signal.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <direct.h>

//$$$ these include files below are watcom c386 / MSDOS ones
#ifdef __DJGPP__
  #include <unistd.h>
  #include <dpmi.h>
  #define _BORLAND_DOS_REGS
#elif defined(__BORLANDC__)
  #include <dir.h>
  #include <direct.h>
  #include <dirent.h>
  #include <dos.h>
  #define _IBM_PCDOS_  // for use of driver
#elif defined(__WATCOMC__)
  #define _IBM_PCDOS_  // for use of driver
  #include <io.h>
  #include <bios.h>
  #include <dos.h>
  #include <conio.h>
  #include <graph.h>
  #include <i86.h>
#else
  #error Compiler not defined
#endif




#include "utility.h"
#include "menu.h"



#define  menualloc(a)  malloc(a)
#define  menufree(a)   free(a)

// internally used data type
// pull down menu stack
typedef struct {
    menutype  * entry;              /* menu entry pointer */
    mword items;                    /* menu entry numbers */
    mword xstart, ystart;           /* menu start location */
    mword width, height;            /* menu display area width height   */
    mword * buffer;                    /* buffer   pointer */
 }   stacktype;

// internally used data type
// menu item's status
typedef struct {
    int message;
    int status;
}  menu_status_type;



#define MSG_QMAX  16   // maximum size of message queue

#define memdisploc       25  // where to display memory status (scrymax, scrxmax-it)
#define busydisploc  12  // where to display busy message (scrymax, scrxmax-it)

#define D_DIRS   1  // dirlist type is directory
#define D_FILES  2  // dirlist type is file
#define D_DRIVES 4  // dirlist type is disk drive
#define D_OTHERS 8  // dirlist type is other

// file list types
typedef struct {
    char attr;
    int  size;
    char *name;
   } _myfs;

#define MAXDIRS   4096 // maximum number of directories
// these atr internal variables
static char **__prs;
static _myfs *__flst;
static int __files;
static char __homes[512];

stacktype * _menu_stack=MNULL;         /* _menu_stack , max level 30 */
mword _menu_stackptr=0;         /* stack pointer */
menu_status_type *_menu_stat=MNULL;
mword _menu_stat_size=0;
mword _menustatus;         /* status word */
mword _menu_message_queue[16];
mword _menu_msg_tail, _menu_msg_header;
char  _menu_tmpbuf[512];

int openmenu(menutype * menu, mword starty, mword startx);
int closemenu(mword level);
int repaint_menu(void);

void movebar(mword curchoise);
void expandmenu(mword curchoise);


//$$$ defines directory sepration char
#define DIRSEPR  '\\'
#define ParentDir ".."
#define CurrentDir "."


//$$$  Below is a section of system dependent routines
// current is DPMI environment (DOS 32 extender)

#define DPMI_INT        0x31
struct meminfo {
    unsigned LargestBlockAvail;
    unsigned MaxUnlockedPage;
    unsigned LargestLockablePage;
    unsigned LinAddrSpace;
    unsigned NumFreePagesAvail;
    unsigned NumPhysicalPagesFree;
    unsigned TotalPhysicalPages;
    unsigned FreeLinAddrSpace;
    unsigned SizeOfPageFile;
    unsigned Reserved[3];
} MemInfo;


void show_memory_free(void)
{
  mword r,c;
#ifdef __DJGPP__
  __dpmi_get_free_memory_information((__dpmi_free_mem_info *)&MemInfo);
#elif defined (__WATCOMC__)
  union REGS regs;
  struct SREGS sregs;
  regs.x.eax = 0x00000500;
  memset( &sregs, 0, sizeof(sregs) );
  sregs.es = FP_SEG( &MemInfo );
  regs.x.edi = FP_OFF( &MemInfo );
  int386x( DPMI_INT, &regs, &regs, &sregs );
#endif
#ifndef __BORLANDC__
  sprintf(_menu_tmpbuf, "Mem %5dkb", MemInfo.LargestBlockAvail/1024);
  crt_grc(&r, &c);
  crt_src(scrymax-1, scrxmax-memdisploc); crt_puts(_menu_tmpbuf, MEMMESS);
  crt_src(r, c);
#endif
}


#ifdef _IBM_PCDOS_
void far critical_error_handler( unsigned deverr,  unsigned errcode, unsigned far *devhdr )
{
  char *mes[10];
  int d_ign = 0, d_ret = 0, d_fail = 0, mnu;
  unsigned btns;
  int k;

  if(deverr&0x8000) { _hardresume(_HARDERR_FAIL);} // Not disk error
  beeper(1);
  mes[0] = "# DOS error ";
  mes[1] = " ";
  switch(errcode&0xff) {
    case 0x00: mes[2] = "* Attempt to write to a write-protected disk "; break;
    case 0x01: mes[2] = "* Unknown unit"; break;
    case 0x02: mes[2] = "* Drive not ready";break;
    case 0x03: mes[2] = "* Unknown command";break;
    case 0x04: mes[2] = "* CRC error in data";break;
    case 0x05: mes[2] = "* Bad drive-request structure length";break;
    case 0x06: mes[2] = "* Seek error";break;
    case 0x07: mes[2] = "* Unknown media type";break;
    case 0x08: mes[2] = "* Sector not found";break;
    case 0x09: mes[2] = "* Printer out of paper";break;
    case 0x0a: mes[2] = "* Write fault";break;
    case 0x0b: mes[2] = "* Read fault";break;
    case 0x0c: mes[2] = "* General failure";break;
    default:   mes[2] = "* Unknown error";break;
  }
  mes[3] = " ";
  mnu=4;
  btns = (deverr&0x3800)>>11;
  if(btns&0x04) { // bit 13==0 indicates "Ignore" response not allowed
     d_ign = 1; mes[mnu++] = "@ Ignore ";
  }
  if(btns&0x02) { // bit 12==0 indicates "Retry" response not allowed
     d_ret = 1; mes[mnu++] = "@ Retry ";
  }
  if(btns&0x01) { // bit 11==0 indicates "Fail" response not allowed
     d_fail = 1; mes[mnu++] = "@ Fail ";
  }
  if(!btns) mes[mnu++] = "@ Abort Program ";
  mes[mnu] = 0;
  k=dialogbox(mes);
  switch(btns) {
    case 0:  // none
      _hardresume(_HARDERR_ABORT);
    case 1:  // fail only
      _hardresume(_HARDERR_FAIL);
    case 2:  // retry only
      _hardresume(_HARDERR_RETRY);
    case 3:  // retry & fail
      if(k == 4) _hardresume(_HARDERR_RETRY);
      else _hardresume(_HARDERR_FAIL);
    case 4:  // ignore only
      _hardresume(_HARDERR_IGNORE);
    case 5:  // ignore & fail
      if(k == 4) _hardresume(_HARDERR_IGNORE);
      else _hardresume(_HARDERR_FAIL);
    case 6: //ignore & retry
      if(k == 4) _hardresume(_HARDERR_IGNORE);
      else _hardresume(_HARDERR_RETRY);
    case 7: // all three
      if(k == 4) _hardresume(_HARDERR_IGNORE);
      if(k == 5) _hardresume(_HARDERR_RETRY);
      else _hardresume(_HARDERR_FAIL);
  }
  _hardresume(_HARDERR_FAIL);
}
#endif // #ifdef _IBM_PCDOS_
//$$$ above is a section of system dependent routines




/********************************************************
*       Initialization of memory area used by menu      *
********************************************************/
int menuinit(int framesel)
  {
   if(!open_screen(framesel)) {
     crt_src(0,0);
     crt_puts("Error initializing screen routines !", 0x9f);
     close_screen();
     return -1;
   }
   _menu_stack = menualloc((int)maxnestlevel*sizeof(stacktype));
   if(_menu_stack == MNULL) {
     crt_src(0,0);
     crt_puts("Can NOT open menu, no free memory!", 0x9f);
     close_screen();
     return(-1);
   }
   _menu_stat=menualloc(maxmenustatus*sizeof(menu_status_type));
//   crt_cls(0,0, scrymax-1, scrxmax-1, BACKGROUND);
   _menu_stackptr = 0; _menu_stat_size=0;
   _menustatus = 0;
   cursor(CURSOR_OFF);
   #ifdef _IBM_PCDOS_
   _harderr( critical_error_handler );
   #endif // #ifdef _IBM_PCDOS_
   return(0);
}


/*********************************************************
*      set free memory and other resources used by menu  *
*********************************************************/
int menuexit()
 {
  mword i;
//  for(i=0; i<scrymax; i++) crt_scroll_down(i,0, scrymax-1, scrxmax-1, BACKGROUND, 1);
//  crt_cls(0,0, scrymax-1, scrxmax-1, BACKGROUND);

  crt_cls(scrymax-1, 0, scrymax-1, scrxmax-1, BACKGROUND); // clear status bar
  if(_menu_stackptr)
     closemenu(1);
//     while(_menu_stackptr-- > 0 )
//         menufree( (_menu_stack+_menu_stackptr)->buffer);
  if( _menu_stack != MNULL) menufree(_menu_stack);
  if( _menu_stat != MNULL) menufree(_menu_stat);
  _menustatus = 0;
  _menu_stat_size = 0;
  cursor(CURSOR_ON);
  close_screen();
  return(0);
 }

  
int raise_message(int message)
{
  if(_menu_message_queue[_menu_msg_header] != 0) return 0; //queue full
  _menu_message_queue[_menu_msg_header] = message;
  _menu_msg_header++;
  if(_menu_msg_header >= MSG_QMAX) _menu_msg_header=0;
  return(message);
}
    
int peek_message(void)
{
  int t;
  if(_menu_message_queue[_menu_msg_tail] == 0) return 0; // empty queue
  t = _menu_message_queue[_menu_msg_tail];
  _menu_message_queue[_menu_msg_tail] = 0;
  _menu_msg_tail++;
  if(_menu_msg_tail >= MSG_QMAX) _menu_msg_tail=0;
  return t;
}

void clear_all_message(void)
{
  int i;
  _menu_msg_header=_menu_msg_tail=0;  // clear message queue
  for(i=0; i<MSG_QMAX; i++) _menu_message_queue[i] = 0;
} 

void __calling(int callbackfunc(int message), int message, int noclose)
{
  mword * savescr;
  mword rtn;
       savescr = menualloc((int)scrxmax*scrymax*2);
       if(savescr != NULL) crt_getblock(savescr, 0,0, scrymax-1, scrxmax-1);
       rtn = callbackfunc(message);
       if(rtn&refresh_screen){
           reset_screen();
           if(savescr != NULL) crt_putblock(savescr, 0,0, scrymax-1, scrxmax-1);
       }
       if(savescr != NULL) menufree(savescr);
       if(rtn) callbackfunc(MENUREPAINT);
       if(rtn&menu_repaint) repaint_menu();
       if(!noclose && (rtn&close_caller_menu)) closemenu(_menu_stackptr);
}

int domenu(menutype * menu, mword starty, mword startx , int callbackfunc(int message))
 {
  stacktype  * curstk;
  choisetype  * menuchoise;
  int key, qkey;
  int curchoise;
  mword tmp, items;
  mword stacktop = _menu_stackptr;

  if(menu == MNULL)  return(-1);
  clear_all_message();
  callbackfunc(MENUINIT);
  callbackfunc(MENUREPAINT);
  if(openmenu(menu, starty, startx) < 0){
     callbackfunc(MENUCLOSE);
     return(-1);
  }
//  callbackfunc(MENUREPAINT);
  crt_src(scrymax-1, 0);
  crt_putsn("", scrxmax, CHOISECOMMENT);
  do   {
     curstk = (_menu_stack+_menu_stackptr-1);
     items = curstk->items;
     curchoise = curstk->entry->choise;
     if(++curstk->entry->choise >= items) curstk->entry->choise = 0;
     movebar(curchoise);
     show_memory_free();
     while((tmp = peek_message()))  __calling(callbackfunc, tmp, 1); //check queue and process it
     while((key=mygetch()) != ESC) {
       switch(key) {
         case UP:
           if(curstk->entry->menutype & mvertical)
                if(--curchoise < 0) curchoise = items-1;
           break;

         case DOWN:
           if(curstk->entry->menutype & mvertical)
                if(++curchoise >= (int)items) curchoise = 0;
           break;

         case LEFT:
           if(curstk->entry->menutype & mvertical)  {
             if(_menu_stackptr >= 2 && !((_menu_stack+_menu_stackptr-2)->entry->menutype & mvertical)) {
                closemenu(_menu_stackptr);
                curstk = (_menu_stack+_menu_stackptr-1);
                items = curstk->items;
                curchoise = curstk->entry->choise;     
                if(--curchoise < 0) curchoise = items-1;
                movebar(curchoise);
                menuchoise = curstk->entry->menulist + curchoise;
                if (menuchoise->menu != MNULL && !(get_menu_status(menuchoise->message)&menu_grey)) {
                    expandmenu(curchoise);
                    goto nextloop;
                }       
             }
           }
           else if(--curchoise < 0) curchoise = items-1;
           break; 
        
         case RIGHT: 
           if(curstk->entry->menutype & mvertical) {
             if(_menu_stackptr >= 2 && !((_menu_stack+_menu_stackptr-2)->entry->menutype & mvertical) ){
                closemenu(_menu_stackptr);
                curstk = (_menu_stack+_menu_stackptr-1);
                items = curstk->items;
                curchoise = curstk->entry->choise;     
                if(++curchoise >= (int)items) curchoise = 0;
                movebar(curchoise);
                menuchoise = curstk->entry->menulist + curchoise;
                if (menuchoise->menu != MNULL && !(get_menu_status(menuchoise->message)&menu_grey)) {
                    expandmenu(curchoise);
                    goto nextloop;
                }       
             }
           }
           else if(++curchoise >= (int)items) curchoise = 0;
           break;

         default:
            if((key&0xff) && (key != ENTER)) {
               key &=0xff;
               if(key >='a' && key <= 'z') key -= 'a'-'A';
               for(tmp = 0;  tmp <items; tmp ++) {
                  if((curstk->entry->menulist + tmp)->qkindex <= 0) continue;
                  qkey= (curstk->entry->menulist + tmp)->qkindex-1;
                  qkey= (curstk->entry->menulist + tmp)->string[qkey];
                  if(qkey >='a' && qkey <= 'z') qkey -= 'a'-'A';
                  if(key == qkey) break;
                }
                if(tmp < items) {
                  curchoise = tmp;
                  movebar(curchoise);
                  key = ENTER;
                }
            }
            if(key == ENTER) {
               menuchoise = curstk->entry->menulist + curchoise;
               if (get_menu_status(menuchoise->message)&menu_grey)  goto nextloop;
               if(curstk->entry->menutype & mselect) { // selector menu
                  curstk->entry->opt = curchoise;
               }   
               if(menuchoise->menu == MNULL) { // only message
                 __calling(callbackfunc, menuchoise->message, 0);
                 goto nextloop;
               }
               else {
                 expandmenu(curchoise);
                 callbackfunc(MENUREPAINT);
                 goto nextloop;
               }
            }
            break;
       }   // switch
       movebar(curchoise);
       show_memory_free();
     }
     if(key == ESC && _menu_stackptr > stacktop+1)
        closemenu(_menu_stackptr);
     nextloop: ;
     while((tmp = peek_message()))  __calling(callbackfunc, tmp, 1); //check queue and process it
  }
  while( stacktop < _menu_stackptr );
  callbackfunc(MENUCLOSE);
  return(0);
} 


int search_menu_status(int message)
{
    register int i;
    if(_menu_stat == MNULL) return -1;
    for(i=0; i<_menu_stat_size && _menu_stat[i].message != message; i++);    
    if(i < _menu_stat_size) return i;
    return -1;
}


int set_menu_status(int message, int stat)
{
   int i;
   if(stat == 0) return 0;
   i=search_menu_status(message);
   if(i >= 0) {
      _menu_stat[i].status |= stat;
      return 1;
   }
   else {
       if(_menu_stat == MNULL || _menu_stat_size == maxmenustatus) return 0;
       _menu_stat[_menu_stat_size].message = message;
       _menu_stat[_menu_stat_size].status  = stat;
       _menu_stat_size ++;
       return 1;
   }
}

int reset_menu_status(int message, int stat)
{
   int i;
   if(stat == 0) return 0;
   i=search_menu_status(message);
   if(i >= 0) {
      if(_menu_stat[i].status & stat) _menu_stat[i].status ^= stat;
      if(_menu_stat[i].status == 0) {
          while(i+1<_menu_stat_size) { _menu_stat[i] = _menu_stat[i+1]; i++; }
          _menu_stat_size--;
      }    
      return 1;
   }
   return 0;
}

int get_menu_status(int message)
{  int i;
   i=search_menu_status(message);
   if(i >= 0) return _menu_stat[i].status;
   else return 0;
}

#define  bnormal   0
#define  bselect   1
#define  bcurrent  2

int show_bar(choisetype *bar, int barlen, mword r, mword c, mword bmode)
{
  int tmp, qktmp;
  int stat;
  char buf[100];
  
  stat = get_menu_status(bar->message);
  if(stat & menu_checked) buf[0] = _checker;
  else if(bmode & bselect) buf[0] = _selector;
  else buf[0] = ' ';
  strcpy(buf+1, bar->string);
  tmp = strlen(buf);
  qktmp=bar->qkindex-1;
  if(qktmp >= 0)
     if(tmp < qktmp) qktmp=bar->qkindex=-1;
  for(;tmp < barlen; tmp++) *(buf+tmp)=' ';
  *(buf+tmp)=0;
  crt_src(r,c);
  if(stat & menu_grey)
       crt_puts(buf, bmode & bcurrent ? CHOOSEDGREY : MENUGREY);
  else {
          crt_puts(buf, bmode & bcurrent ? CHOOSEDTITLE : MENUBACK);
          if(qktmp>=0) {
            crt_src(r, c+1+qktmp);
            crt_putch(bar->string[qktmp], bmode & bcurrent ? QUICKKEYCHOOSED : QUICKKEY);
          }
  }
  return qktmp;
}  

int openmenu(menutype * menu, mword starty, mword startx)
 {
  int items;
  int horlen;
  int tmp, lth;
  int barlen;
  char * str;

  if(menu == MNULL) return(-1);
  if( _menu_stackptr >= maxnestlevel) return(-1);
  //if(!refresh) (_menu_stack + _menu_stackptr)->choise = 0;
  (_menu_stack + _menu_stackptr)->entry  = menu;
  (_menu_stack + _menu_stackptr)->xstart = startx;
  (_menu_stack + _menu_stackptr)->ystart = starty;
  if( menu->menutype & mvertical) { 
     for(items = 0, barlen=0; (menu->menulist + items)->string != NULL; items ++) {
          str = (menu->menulist+items)->string;
          tmp = strfindc(str, ESCAPE_CHAR);
          if(tmp >= 0) {
              (menu->menulist+items)->qkindex = tmp+1;
              do { str[tmp] = str[tmp+1]; tmp++; } while(str[tmp]); //kill escape_char
          }       
          else if ((menu->menulist+items)->qkindex <= 0) (menu->menulist+items)->qkindex = -1;
          if((tmp = strlen(str)) > barlen) barlen = tmp;
     }
     menu->barlen = barlen;
     menu->items  = items;
     if(!items) return(-1);
     (_menu_stack + _menu_stackptr)->items  = items;
     (_menu_stack + _menu_stackptr)->width  = barlen + 2 + (mframe & menu->menutype ? 2:0)
                        + (mshadow & menu->menutype ? 1:0);
     (_menu_stack + _menu_stackptr)->height = items  + (mframe & menu->menutype ? 2:0)
                        + (mshadow & menu->menutype ? 1:0);

     if( starty+(_menu_stack + _menu_stackptr)->height > scrymax) {
         starty = scrymax - (_menu_stack + _menu_stackptr)->height;
         if(starty > scrymax) starty = 0;
         (_menu_stack + _menu_stackptr)->ystart = starty;
     }         

     if( startx+(_menu_stack + _menu_stackptr)->width > scrxmax) {
         startx = scrxmax - (_menu_stack + _menu_stackptr)->width;
         if(startx > scrxmax) startx = 0;
         (_menu_stack + _menu_stackptr)->xstart = startx;
     }         
     
     (_menu_stack + _menu_stackptr)->buffer = menualloc((long)(_menu_stack + _menu_stackptr)->width * (_menu_stack + _menu_stackptr)->height*2);
     if((_menu_stack + _menu_stackptr)->buffer == NULL) return(-1);
     crt_getblock((_menu_stack + _menu_stackptr)->buffer ,
                  starty,            startx, 
                  starty+(_menu_stack + _menu_stackptr)->height-1, 
                  startx+(_menu_stack + _menu_stackptr)->width-1        );
     for(; items > 0; items--) {
        if(menu->menutype & mselect && items-1 == menu->opt) 
           show_bar(menu->menulist+items-1, barlen+2,  
                 starty+items - (mframe & menu->menutype ? 0:1),
                 startx+(mframe & menu->menutype ? 1:0),        bselect);
        else 
           show_bar(menu->menulist+items-1, barlen+2,  
                 starty+items - (mframe & menu->menutype ? 0:1),
                 startx+(mframe & menu->menutype ? 1:0),        bnormal);
     } // for
  }   // mvertical
  else{
     for(items = 0, horlen=0; (menu->menulist + items)->string != NULL; items ++) {
          str = (menu->menulist+items)->string;
          tmp = strfindc(str, ESCAPE_CHAR);
          if(tmp >= 0) {
              (menu->menulist+items)->qkindex = tmp+1;
              do { str[tmp] = str[tmp+1]; tmp++; } while(str[tmp]); //kill escape_char
          }       
          else if ((menu->menulist+items)->qkindex <= 0) (menu->menulist+items)->qkindex = -1;
          horlen += strlen(str)+2;
     }
     menu->barlen = horlen;
     menu->items  = items;
     if(!items) return(-1);
     *_menu_tmpbuf = 0; 
     if(menu->menutype & mfullwidth) {
         horlen = scrxmax - (mframe & menu->menutype ? 2:0) - (mshadow & menu->menutype ? 1:0);
         for(tmp = 0; tmp <horlen-menu->barlen; tmp ++) _menu_tmpbuf[tmp]=' ';
         _menu_tmpbuf[tmp] = 0;
     }
     (_menu_stack + _menu_stackptr)->items  = items;
     (_menu_stack + _menu_stackptr)->width  = horlen + (mframe & menu->menutype ? 2:0)  + (mshadow & menu->menutype ? 1:0);
     (_menu_stack + _menu_stackptr)->height = 1      + (mframe & menu->menutype ? 2:0)  + (mshadow & menu->menutype ? 1:0);
     (_menu_stack + _menu_stackptr)->buffer = menualloc((long)(_menu_stack + _menu_stackptr)->width * (_menu_stack + _menu_stackptr)->height*2);
     if(starty+(_menu_stack + _menu_stackptr)->height > scrymax) {
         starty = scrymax - (_menu_stack + _menu_stackptr)->height;
         if(starty > scrymax) starty = 0;
         (_menu_stack + _menu_stackptr)->ystart = starty;
     }
     if( startx+(_menu_stack + _menu_stackptr)->width > scrxmax) {
         startx = scrxmax - (_menu_stack + _menu_stackptr)->width;
         if(startx > scrxmax) startx = 0;
         (_menu_stack + _menu_stackptr)->xstart = startx;
     }         
     if((_menu_stack + _menu_stackptr)->buffer == NULL) return(-1);
     crt_getblock((_menu_stack + _menu_stackptr)->buffer ,
                  starty,            startx, 
                  starty+(_menu_stack + _menu_stackptr)->height-1, 
                  startx+(_menu_stack + _menu_stackptr)->width-1        );
     for(tmp=0, lth=0; tmp<items; tmp++) {
        barlen = strlen((menu->menulist+tmp)->string)+2;
        if(menu->menutype & mselect && tmp == menu->opt) 
           show_bar(menu->menulist+tmp, barlen,  
                 starty+(mframe & menu->menutype ? 1:0), 
                 startx+(mframe & menu->menutype ? 1:0)+lth, bselect);
        else
           show_bar(menu->menulist+tmp, barlen,  
                 starty+(mframe & menu->menutype ? 1:0), 
                 startx+(mframe & menu->menutype ? 1:0)+lth, bnormal);
        lth += barlen;
     }
     if(*_menu_tmpbuf) { // full length menu required
        crt_src(starty+(mframe & menu->menutype ? 1:0), startx+(mframe & menu->menutype ? 1:0)+lth);
        crt_puts(_menu_tmpbuf, MENUBACK);
     }    
  }
  
  if( menu->menutype & mshadow )                   
     shadow(starty, startx, starty+(_menu_stack + _menu_stackptr)->height-1,startx+(_menu_stack + _menu_stackptr)->width-1, SHADOW);
  if( menu->menutype & mframe)
     frame(starty,           startx,
           starty+(_menu_stack + _menu_stackptr)->height-1-(mshadow & menu->menutype ? 1:0),
           startx+(_menu_stack + _menu_stackptr)->width -1-(mshadow & menu->menutype ? 1:0),
           MENUEDGE);

  _menu_stackptr++;
  return(0);
}



int closemenu(mword     stacktop )
 {
  if( stacktop <= 0) return(-1);
  for(;_menu_stackptr>=stacktop; _menu_stackptr --) {
        crt_putblock((_menu_stack + _menu_stackptr - 1)->buffer ,
                   (_menu_stack + _menu_stackptr - 1)->ystart,
                   (_menu_stack + _menu_stackptr - 1)->xstart, 
                   (_menu_stack + _menu_stackptr - 1)->ystart+(_menu_stack + _menu_stackptr -1)->height-1, 
                   (_menu_stack + _menu_stackptr - 1)->xstart+(_menu_stack + _menu_stackptr -1)->width-1        );
        menufree((_menu_stack + _menu_stackptr - 1)->buffer);
    }
  return 0;
 }

int repaint_menu(void)
 {
  mword stacktop=_menu_stackptr;

  if( _menu_stackptr <= 0) return(-1);
  for(;_menu_stackptr>0; _menu_stackptr --) {
        crt_putblock((_menu_stack + _menu_stackptr - 1)->buffer ,
                   (_menu_stack + _menu_stackptr - 1)->ystart,
                   (_menu_stack + _menu_stackptr - 1)->xstart,
                   (_menu_stack + _menu_stackptr - 1)->ystart+(_menu_stack + _menu_stackptr -1)->height-1,
                   (_menu_stack + _menu_stackptr - 1)->xstart+(_menu_stack + _menu_stackptr -1)->width-1        );
        menufree((_menu_stack + _menu_stackptr - 1)->buffer);
    }
  openmenu(_menu_stack->entry, _menu_stack->ystart, _menu_stack->xstart);
  movebar((_menu_stack+_menu_stackptr-1)->entry->choise);
  while(_menu_stackptr<stacktop) {
    expandmenu((_menu_stack+_menu_stackptr-1)->entry->choise);
    movebar((_menu_stack+_menu_stackptr-1)->entry->choise);
  }
  return 0;
 }


    
void movebar(mword curchoise)
 { 
  stacktype  * curstk;
  mword xstart, ystart, width;
  mword tmp, cnt;
        curstk = (_menu_stack + _menu_stackptr -1);
        width = curstk->width - (mframe & curstk->entry->menutype ? 2:0)
             - (mshadow & curstk->entry->menutype ? 1:0);
        xstart = curstk->xstart+(mframe & curstk->entry->menutype ? 1:0);
        ystart = curstk->ystart+(mframe & curstk->entry->menutype ? 1:0);
/*      if(curchoise != curstk->entry->choise)
*/      {
          if(curstk->entry->menutype & mvertical)  {
             if(curstk->entry->menutype & mselect 
               && curstk->entry->choise == curstk->entry->opt)
                show_bar(curstk->entry->menulist + curstk->entry->choise, width,
                      ystart+curstk->entry->choise, xstart, bselect);
             else
                show_bar(curstk->entry->menulist + curstk->entry->choise, width,
                      ystart+curstk->entry->choise, xstart, bnormal);
          }
          else  {
             for(tmp = 0, cnt = 0; tmp < curstk->entry->choise; tmp++)
                cnt += strlen((curstk->entry->menulist + tmp)->string)+2;
             if(curstk->entry->menutype & mselect 
               && tmp == curstk->entry->opt)
                show_bar(curstk->entry->menulist + tmp, 
                      strlen((curstk->entry->menulist + tmp)->string)+2,
                      ystart, xstart+cnt, bselect);
             else          
                show_bar(curstk->entry->menulist + tmp,
                      strlen((curstk->entry->menulist + tmp)->string)+2,
                      ystart, xstart+cnt, bnormal);
          }
          curstk->entry->choise = curchoise;
          if(curstk->entry->menutype & mvertical) {
            if(curstk->entry->menutype & mselect 
              && curchoise == curstk->entry->opt)
                show_bar(curstk->entry->menulist + curchoise, width,
                     ystart+curchoise, xstart, bselect|bcurrent);
            else     
                show_bar(curstk->entry->menulist + curchoise, width,
                     ystart+curchoise, xstart, bcurrent);
          }
          else  {
            for(tmp = 0, cnt = 0; tmp < curchoise; tmp++)
               cnt += strlen((curstk->entry->menulist + tmp)->string)+2;
            if(curstk->entry->menutype & mselect 
              && curchoise == curstk->entry->opt)
               show_bar(curstk->entry->menulist + curchoise, 
                     strlen((curstk->entry->menulist + tmp)->string)+2,
                     ystart, xstart+cnt, bselect|bcurrent);
            else
               show_bar(curstk->entry->menulist + curchoise, 
                     strlen((curstk->entry->menulist + tmp)->string)+2,
                     ystart, xstart+cnt, bcurrent);
          }
          crt_src(scrymax-1, 0);
          crt_putsn((curstk->entry->menulist + curchoise)->comment, scrxmax-memdisploc, CHOISECOMMENT);
        }
  }

void expandmenu(mword curchoise)
{
  stacktype * curstk;
  mword cnt, tmp;       
  mword xstart, ystart, width;
  choisetype *     menuchoise;

         curstk = (_menu_stack + _menu_stackptr -1);
         menuchoise = curstk->entry->menulist + curchoise;
         width = curstk->width - (mframe & curstk->entry->menutype ? 2:0)
             - (mshadow & curstk->entry->menutype ? 1:0);
         xstart = curstk->xstart+(mframe & curstk->entry->menutype ? 1:0);
         ystart = curstk->ystart+(mframe & curstk->entry->menutype ? 1:0);

         if (curstk->entry->menutype & mvertical) {
           ystart += curchoise+1;
           xstart ++;
         }
         else {
           ystart +=1;    
           for(tmp = 0, cnt = 0; tmp < curchoise; tmp++)
             cnt += strlen((curstk->entry->menulist + tmp)->string)+2;
           crt_src(ystart, xstart+cnt);
           xstart += cnt;
         }
         openmenu(menuchoise->menu, ystart, xstart);
}
                           


void menu_busy(short busy)
{
    static mword busy_sto[12];
    static short busy_on=0;
    static short cnt=0;
    static char runner[4]={'|', '/', '-', '\\'}; 
    char bsy[13];
    mword r,c;
    if(busy > 0) {
        if(!busy_on) {   
           crt_getblock(busy_sto, scrymax-1, scrxmax-busydisploc, scrymax-1, scrxmax-1);
        }   
        cnt++;
        crt_grc(&r, &c);
        crt_src(scrymax-1, scrxmax-12); 
        if(busy > 100) 
          sprintf(bsy, " Busy ... %c", runner[cnt&0x03]);
        else
          sprintf(bsy, " %3d%% done%c", busy, runner[cnt&0x03]);
        crt_puts(bsy, BUSYMESS);
        crt_src(r, c);
        busy_on = 1;
    }
    else {
        if(!busy_on) return;
        crt_putblock(busy_sto, scrymax-1, scrxmax-busydisploc, scrymax-1, scrxmax-1);
        busy_on = 0;
    }
    return;
}





void putnotice(mword row, char * message, mword att)
 {mword col;
  col = (scrxmax-strlen(message))/2;
  crt_src(row,col);
  crt_puts(message, att);
  shadow(row, col, row+1, col+strlen(message), SHADOW);
 }


int dialogbox(char * message[])
 {
  mword row, col, len,linb, i,j, num, buttonind, sumbutton=0, mb, lin, bscol;
  mword * tmpsave; 
  int button[20], bcol[20], brow[20];
  char *title;
  int boxst;

  buttonind = 0; sumbutton = 0; mb = 0;
  if(message[0] == NULL) return 0;
  len = 0;  boxst=0;  title = NULL;
  if(*(message[0]) == '#') {
    title = message[0]+1;
    len = strlen(title);
    if(len == 0) title = NULL;
    boxst = 1;
  }    
  for(i = boxst, lin = 0; message[i] != NULL; i++)
   {
     if(*(message[i]) == '@')
        {
        button[buttonind] =i;
        sumbutton += strlen(message[i])+1;
        buttonind++; mb = 1;
        }
     else
        {
        if( mb )
         {
           lin ++;
           if( sumbutton > len ) len = sumbutton;
           mb = 0;  sumbutton = 0 ;
         }
        if( len < strlen(message[i])) len = strlen(message[i]);
        lin ++;
        }
   }
  if( sumbutton > len ) len = sumbutton;
  len += 3;
  lin++;
  linb = lin;
  num = i;
  if(! num) return(0);
  if((tmpsave=menualloc((long)(lin+4)*(len+3)*2)) == NULL) return(-1);
  col = (scrxmax - len)/2 - 1 ;
  row = (scrymax - lin)/2 - 2 ;
  crt_getblock(tmpsave, row, col, row+lin+3, col+len+2);
  frame(row, col, row+lin+2, col+len+1, DIALOGEDGE);
  if(title) {
     crt_src(row, col+(len-strlen(title))/2+1);
     crt_puts(title, DIALOGTITLE);
  }   
  crt_cls(row+1, col+1, row+lin+1, col+len, DIALOGBACK);
  shadow(row, col, row+lin+3, col+len+2,SHADOW);
  for(i = boxst, lin = 0, buttonind = 0; i < num; i++, lin++)
   {
    if(*(message[i]) == '@')
     {
         for(j = i, sumbutton = 0, mb = 0; *(message[j]) == '@'; j++, mb++)
          sumbutton += strlen(message[j])+1;
         bscol = col+(len-sumbutton)/2+2;
         for(j = 0, sumbutton = 0; j < mb ; j++, i++, buttonind++)
         {
           brow[buttonind] = row+lin+1; bcol[buttonind] = bscol+sumbutton;
           crt_src(row+lin+1, bscol+sumbutton);
           crt_puts(message[i]+1, BUTTONUP);
           sumbutton += strlen(message[i])+1;
         }
         i--;

     }
    else if(*(message[i]) == '*')
     { crt_src(row+lin+1, col+1);
         crt_puts(message[i]+1, ERRORMESS);
     }
    else
     { crt_src(row+lin+1, col+(len-strlen(message[i]))/2+1);
         crt_puts(message[i], DIALOGTEXT);
     }
   }    

  if(buttonind) {
    sumbutton = 0;
    crt_src(brow[sumbutton], bcol[sumbutton]);
    crt_puts(message[button[sumbutton]]+1, BUTTONDOWN);
  }
  while((i=mygetch())!=ENTER && i != ESC)
   if(buttonind)     /* has button */
     {
        switch(i)
         {
         case UP  :
         case LEFT:
            crt_src(brow[sumbutton], bcol[sumbutton]);
            crt_puts(message[button[sumbutton]]+1, BUTTONUP);
            if(sumbutton == 0) sumbutton = buttonind-1;
            else sumbutton--;
            break;
         case RIGHT:
         case DOWN :
         case TAB  :
            crt_src(brow[sumbutton], bcol[sumbutton]);
            crt_puts(message[button[sumbutton]]+1, BUTTONUP);
            if(sumbutton == buttonind-1) sumbutton = 0;
            else sumbutton++;
            break;
         }
        crt_src(brow[sumbutton], bcol[sumbutton]);
        crt_puts(message[button[sumbutton]]+1, BUTTONDOWN);
     }
  if(i == ESC) sumbutton = 0;
  crt_putblock(tmpsave, row, col, row+linb+3, col+len+2);
  menufree(tmpsave);
  if(buttonind)  return(button[sumbutton]);
  else return(0);
 }



int selectlist(char **list, int starty, int startx , int ysz, int xsz, int callbackfunc(int message), char *title)
 {
  mword * savescr;
  int key, curc, items;
  int i, st, lastc, lasts, rtn;
  
  for(items=0; list[items] != NULL; items++);  // count list number
  if (!items) {
    rtn = callbackfunc(0);
    return 0;
  }
  if(ysz > items) ysz = items;
  if(ysz < 2) ysz = 2;
  if (starty+ysz+2 >= scrymax) {
     starty = scrymax-ysz-2;
     if(starty < 0) { 
        starty = 0;
        ysz = scrymax-2;
      }
   }
  if (startx+xsz+2 >= scrxmax) {
     startx = scrxmax-xsz-2;
     if(startx < 0) {
        startx = 0;
        xsz = scrxmax-2;
      }
   }
  savescr = menualloc((int)(xsz+2)*(ysz+2)*2);
  if(savescr != NULL) crt_getblock(savescr, starty, startx, starty+ysz+1, startx+xsz+1);
  frame(starty, startx, starty+ysz+1, startx+xsz+1, LISTEDGE);
  if(title != NULL) {
    i = strlen(title);
    if(xsz < i) i = xsz;
    crt_src(starty, startx + 1 + (xsz-i)/2);
    crt_putsn(title, i, LISTEDGE);
  }
  st = 0; curc = 0;  lastc = 0; lasts = 0;
  for(i=0; i<ysz; i++){
    crt_src(starty+1+i, startx+1); 
    if(i+st<items) crt_putsn(list[i+st], xsz, LISTTEXT);
    else  crt_putsn("", xsz, LISTTEXT);
  }
  drawscrollbar(starty+1, startx+xsz+1, ysz-2, vertical_scroll_bar, LISTEDGE);
  setscrollbar(starty+1, startx+xsz+1, ysz-2, vertical_scroll_bar, LISTEDGE, 0, 0);
  crt_src(starty+1+curc-st, startx+1);
  crt_putsn(list[curc], xsz, LISTSEL);
  rtn = callbackfunc(curc+1);
  while(curc >= 0 && rtn) {
    if(lastc != curc) {
        crt_src(starty+1+lastc-st, startx+1); 
        if(lastc <items) crt_putsn(list[lastc], xsz, LISTTEXT);
        else  crt_putsn("", xsz, LISTTEXT);
        crt_src(starty+1+curc-st, startx+1);
        if(curc <items) crt_putsn(list[curc], xsz, LISTSEL);
        i = (ysz-2)*curc/items;
        setscrollbar(starty+1, startx+xsz+1, ysz-2, vertical_scroll_bar, LISTEDGE, lasts, i);
        lasts=i; lastc = curc;
    }
    key = mygetch();
    switch(key) {
      case RIGHT:
      case DOWN :
        if(curc == items-1) {beeper(4);}
        else { 
          curc++;
          if(curc-st >= ysz){  // scroll list up
            crt_scroll_up(starty+1, startx+1, starty+ysz, startx+xsz, LISTTEXT, 1);
            st++;
          }
        }
        rtn = callbackfunc(curc+1);
        break;
/*      case PGDN:
        if(curc == items-1) {beeper(4);}
        else { 
          curc+=ysz-2;
          if(curc >= items) curc == items-1;
          if(curc-st >= ysz){  // scroll list up
            crt_scroll_up(starty+1, startx+1, starty+ysz, startx+xsz, LISTTEXT, curc-st-ysz+1);
            st+=curc-st-ysz+1;
          }
        }
        rtn = callbackfunc(curc+1);
        break;

      case PGUP:
*/      case UP  :
      case LEFT:
        if(curc == 0) {beeper(4);}
        else { 
          curc--;
          if(curc-st < 0){  // scroll list up
            crt_scroll_down(starty+1, startx+1, starty+ysz, startx+xsz, LISTTEXT,1);
            st--;
          }
        }
        rtn = callbackfunc(curc+1);
        break;
      case ENTER:
      case SPACEBAR:
        rtn = callbackfunc(-(curc+1));
        break;
      case ESC:
        rtn = callbackfunc(0);
        curc = -1;
        break;
    }
  };
  if(savescr != NULL) {
     crt_putblock(savescr, starty, startx, starty+ysz+1, startx+xsz+1);
     menufree(savescr);
  }  
  return(curc+1);
} 

   
int get_string(char *str, int lth, char *comment, int keeps)
{
   mword * tmpsave;

   if((tmpsave=menualloc((long)scrxmax*6)) != NULL)
      crt_getblock(tmpsave, scrymax-4, 0, scrymax-2, scrxmax-1);
   crt_cls(scrymax-4, 0, scrymax-2, scrxmax-1, BACKGROUND);
   frame(scrymax-4, 0, scrymax-2, scrxmax-1, DIALOGEDGE);
   crt_src(scrymax-3,1);
   crt_puts(comment, 0x1b);
   if(!keeps) str[0] = 0;
   if(myeditline(str, 0x1a, lth, scrxmax-strlen(comment)-2) == ESC)
    {
     crt_putblock(tmpsave, scrymax-4, 0, scrymax-2, scrxmax-1);
     if(tmpsave) menufree(tmpsave);
     return(0);
    }
   crt_putblock(tmpsave, scrymax-4, 0, scrymax-2, scrxmax-1);
   if(tmpsave) menufree(tmpsave);
   return(strlen(str));
}


int getfilename(char * fname, char * ext, char *comment)
 {
   if(get_string(fname, 60, comment, 1)) {
       append_extension(fname, ext);  
       return(strlen(fname));
   }
   return 0;
 }


int getnumber(char *comment)
{
  char bf[32];
  int i;
  i=get_string(bf, 30, comment, 0);
  if(!i) return 0;
  return atoi(bf);
} 

double getfloat(char *comment)
{
  char bf[32];
  int i;
  i=get_string(bf, 30, comment, 0);
  if(!i) return 0.0;
  return atof(bf);
} 


void dir_err(int err)
{
  char *mes[10];
  
  mes[0] = "# Directory Access Error ";
  mes[1] = " ";
  if(err>0)
    switch(errno) {
      case ERANGE:
        mes[2] = "* Directory name is too long. ";
        break;
      case ENOENT:
        mes[2] = "* Directory not found. ";
        break;
      case EACCES:
        mes[2] = "* Permission denied. ";
        break;
      default:
        mes[2] = "* Error. ";
    }
  else 
    switch(err) {
      case 1:
      default:
        mes[2] = "* Error getting file name";
        break;
    }
  mes[3] = " ";
  mes[4] = "@ Ok ";
  mes[5] = NULL;
  dialogbox(mes);
  return;
}

int dircallback(int idx)
{
    char tbf[256];
    int m = strlen(__homes);
    int rtn=0;
    
    if( m > 0 && __homes[m-1] == DIRSEPR) m = 1;
    else m = 0;
    crt_src(scrymax-1, 0);
    if(idx == 0)  tbf[0] = 0;
    else {
      if(idx < 0) { rtn = 0; idx = -idx-1;}
      else {idx--; rtn = 1; }
      if(__flst[idx].attr == D_DRIVES)
        sprintf(tbf, "Disk Drive '%c:'", __flst[idx].name[1]);
      else if(m)
        if(__flst[idx].attr == D_FILES)
           sprintf(tbf, "%s%s  %dbytes", __homes, __flst[idx].name, __flst[idx].size);
        else
           sprintf(tbf, "%s%s  <Directory>", __homes, __flst[idx].name);
      else if(__flst[idx].attr == D_FILES)
           sprintf(tbf, "%s%c%s  %dbytes", __homes, DIRSEPR, __flst[idx].name, __flst[idx].size);
      else
           sprintf(tbf, "%s%c%s  <Directory>", __homes, DIRSEPR, __flst[idx].name);
    }
    crt_putsn(tbf, scrxmax-memdisploc,   0x70);
    return rtn;
}

int dirstrip(char *dirs)
{int i,j,k, s;
 k = strlen(dirs);
 for(i=0,s=0; i<k; i++) if(dirs[i]==DIRSEPR) {j=i; s++;}

 if(s>1) {
   dirs[j] = 0; return 1;
 }
 else if(s==1 && j < k-1) {
   dirs[j+1]=0; return 1;
 }
 
// if(s) { dirs[j] = 0; return 1; }
 return 0;
}


//$$$ this is for systems with drive concepts
#ifdef _IBM_PCDOS_
void __append_drive_to_list(void)
{
  unsigned dv, ndv, i, k;

  _dos_getdrive(&dv);
  _dos_setdrive(dv, &ndv);
  for(i=3; i<=ndv; i++) {
    _dos_setdrive(i, &ndv);
    _dos_getdrive(&k);
    if(i != k) ndv = i-1;
  }
  _dos_setdrive(dv, &k);

  for(i=1; i<=ndv; i++) {
    __prs[__files] = malloc(5);
    if(__prs[__files] == NULL) return;
    __prs[__files][0] = '[';
    __prs[__files][1] = i-1+'A';
    __prs[__files][2] = ':';
    __prs[__files][3] = ']';
    __prs[__files][4] = 0;
    __flst[__files].attr = D_DRIVES;
    __flst[__files].name = __prs[__files]; //??
    __files ++;
  }
}
#endif //#ifdef _IBM_PCDOS_

#ifdef __BORLANDC__
// $$$
static struct ffblk _fblk;

// $$$
int dir_getfirst(char *path, char *name, long *size)
{
  if(findfirst(path, &_fblk, FA_RDONLY | FA_HIDDEN | FA_SYSTEM | FA_DIREC)) return 0; // failed
  strcpy(name, _fblk.ff_name);
  *size = _fblk.ff_fsize;
  if(_fblk.ff_attrib & FA_DIREC) return D_DIRS;
  else if(_fblk.ff_attrib & FA_LABEL)) return D_OTHERS;
  else return D_FILES;
}

// $$$
int dir_getnext(char *name, long *size)
{
  if(findnext(&_fblk)) return 0;
  strcpy(name, _fblk.ff_name);
  *size = _fblk.ff_fsize;
  if(_fblk.ff_attrib & FA_DIREC) return D_DIRS;
  else if(_fblk.ff_attrib & FA_LABEL) return D_OTHERS;
  else return D_FILES;
}

void dir_close(void) {}

#elif defined(__WATCOMC__)
// $$$
static struct _find_t _fblk;


// $$$
int dir_getfirst(char *path, char *name, long *size)
{
  if(_dos_findfirst(path, _A_NORMAL | _A_RDONLY |
                   _A_SYSTEM | _A_HIDDEN | _A_SUBDIR, &_fblk)) return 0; // failed
  strcpy(name, _fblk.name);
  *size = _fblk.size;
  if(_fblk.attrib & _A_SUBDIR) return D_DIRS;
  else if(_fblk.attrib & (_A_VOLID)) return D_OTHERS;
  else return D_FILES;
}

// $$$
int dir_getnext(char *name, long *size)
{
  if(_dos_findnext(&_fblk)) return 0;
  strcpy(name, _fblk.name);
  *size = _fblk.size;
  if(_fblk.attrib & _A_SUBDIR) return D_DIRS;
  else if(_fblk.attrib & _A_VOLID) return D_OTHERS;
  else return D_FILES;
}

void dir_close(void)
{
}

#else
  #error  Compiler not defined
#endif

// $$$
char *dir_getcwd(char *buf, int len)
{
  return getcwd(buf, len);
}

// $$$
int dir_chdir(char *dir)
{
  return chdir(dir);
}



int dirsel(char *defmasks, char *res, char *title, int onlydir)
{
  int dir_rtn;
  long fsize;
  char fname[256];
  char hmask[512];
  char local_dir[256];
  char *masks;
  char *tpr;
  int i, k;
  int tlp;
  #ifdef _IBM_PCDOS_
  unsigned int old_drv, dummy;
  #endif

  masks = (defmasks == NULL || defmasks[0] == 0) ? "*.*" : defmasks;

  __prs = malloc(sizeof(char *)*MAXDIRS);
  if(__prs == NULL) return -1;
  __flst = malloc(sizeof(_myfs)*MAXDIRS);
  if(__flst == NULL) {
     free(__prs);
     return -1;
  }
  tpr = dir_getcwd(__homes, 256);
  if(tpr == NULL) {
     dir_err(-1);
     free(__flst); free(__prs);
     return -1;
  }
  strcpy(local_dir, __homes);
  
  #ifdef _IBM_PCDOS_
  _dos_getdrive(&old_drv);
  #endif
    
  do {
    if(dir_chdir(__homes)) {
       dir_err(-1);
       free(__flst); free(__prs);
       dir_chdir(local_dir);
       #ifdef _IBM_PCDOS_
//       _dos_setdrive(old_drv, &dummy);
       #endif
       return -1;
    }
    
    strcpy(hmask, __homes);
    k = strlen(hmask);
    if(k > 0 && hmask[k-1] != DIRSEPR) {  // we need a dirsepr now
      hmask[k] = DIRSEPR;
      hmask[k+1] = 0;
    }
    //$$$ DOS ONLY
    strcat(hmask, "*.*");
    dir_rtn = dir_getfirst(hmask, fname, &fsize);
    if(!dir_rtn)  {
       dir_err(-1);
       free(__flst); free(__prs);
       dir_chdir(local_dir);
       #ifdef _IBM_PCDOS_
//       _dos_setdrive(old_drv, &dummy);
       #endif
       return -1;
    }
    strcpy(hmask, __homes);
    if(onlydir) {
      __prs[0] = ". [Select Current Directory]";
      __flst[0].attr = D_DIRS;
      __flst[0].name = __prs[0];  //??
      __files = 1;
    }
    else __files = 0;
    if(dirstrip(hmask) > 0) {
      __prs[__files] = ParentDir;
      __flst[__files].attr = D_DIRS;
      __flst[__files].name = __prs[__files];  //??
      __files++;
    }
    tlp = __files;
    for(;;) {
        if(dir_rtn == D_DIRS && fname[0] != '.') {
             i = strlen(fname);
             __prs[__files] = malloc(i+2);
             if(__prs[__files] == NULL) {
                for(i=__files-1; i>=tlp; i--) free(__prs[i]);
                dir_chdir(local_dir);
                #ifdef _IBM_PCDOS_
//                _dos_setdrive(old_drv, &dummy);
                #endif
                free(__flst); free(__prs); return -1;
             }
             strcpy(__prs[__files], fname);
             __prs[__files][i] = DIRSEPR;  __prs[__files][i+1] = 0;
             __flst[__files].attr = D_DIRS;
             __flst[__files].name = __prs[__files]; //??
             __files ++;
        }
        dir_rtn = dir_getnext(fname, &fsize);
        if(!dir_rtn) break;
    }
    strcpy(hmask, __homes);
    if(strlen(masks)) {
      k = strlen(hmask);
      if(k > 0 && hmask[k-1] != DIRSEPR) {  // we need a dirsepr now
        hmask[k] = DIRSEPR;
        hmask[k+1] = 0;
      }
      strcat(hmask, masks);
    }
    dir_close();
    if(!onlydir) {
      dir_rtn = dir_getfirst(hmask, fname, &fsize);
      if(!dir_rtn && __files == 0)  {
         dir_err(-1);
         for(i=__files-1; i>=tlp; i--) free(__prs[i]);
         free(__flst); free(__prs);
         dir_chdir(local_dir);
         #ifdef _IBM_PCDOS_
//         _dos_setdrive(old_drv, &dummy);
         #endif
         return -1;
      }
      for(;;) {
        if(dir_rtn == D_FILES) {
             i = strlen(fname);
             __prs[__files] = malloc(i+1);
             if(__prs[__files] == NULL) {
                for(i=__files-1; i>=tlp; i--) free(__prs[i]);
                dir_chdir(local_dir);
                #ifdef _IBM_PCDOS_
//                _dos_setdrive(old_drv, &dummy);
                #endif
                free(__flst); free(__prs); return -1;
             }
             strcpy(__prs[__files], fname);
             __flst[__files].attr = D_FILES;
             __flst[__files].size = fsize;
             __flst[__files].name = __prs[__files]; //??
             __files ++;
        }
        dir_rtn = dir_getnext(fname, &fsize);
        if(!dir_rtn) break;
      }
    }
    dir_close();
    #ifdef _IBM_PCDOS_
    __append_drive_to_list();
    #endif // #ifdef _IBM_PCDOS_
    __prs[__files] = NULL;
select_dir:
    k = selectlist(__prs, scrymax/5, scrxmax/6, scrymax*3/5, scrxmax*2/3,
                   dircallback, title==NULL ? __homes : title);
    if(k == 0) {
       for(i=__files-1; i>=tlp; i--) free(__prs[i]);
       free(__flst); free(__prs);
       dir_chdir(local_dir);
       #ifdef _IBM_PCDOS_
//       _dos_setdrive(old_drv, &dummy);
       #endif
       return 0;
    }
    k--;
    if(k == 0 && onlydir) {
       strcpy(res, __homes);
       for(i=__files-1; i>=tlp; i--) free(__prs[i]);
       free(__flst); free(__prs);
       dir_chdir(local_dir);
       #ifdef _IBM_PCDOS_
//       _dos_setdrive(old_drv, &dummy);
       #endif
       return 1;
    }
    if(__flst[k].attr == D_FILES) {
       strcpy(res, __homes);
       i = strlen(res);
       if(res[i-1] != DIRSEPR) {
         res[i] = DIRSEPR;
         res[i+1] = 0;
       }
       strcat(res, __flst[k].name);
       for(i=__files-1; i>=tlp; i--) free(__prs[i]);
       free(__flst); free(__prs);
       dir_chdir(local_dir);
       #ifdef _IBM_PCDOS_
//       _dos_setdrive(old_drv, &dummy);
       #endif
       return 1;
    }
    // $$$
    #ifdef _IBM_PCDOS_
    else if(__flst[k].attr == D_DRIVES) {
//       unsigned dummy, odrv;
//       _dos_getdrive(&odrv);
       _dos_setdrive(__flst[k].name[1]+1-'A', &dummy);
       _dos_getdrive(&dummy);
       if(dummy != (unsigned)(__flst[k].name[1]+1-'A')) {
          _dos_setdrive(old_drv, &dummy);
          beeper(3); goto select_dir;
       }
       tpr = dir_getcwd(__homes, 256);
       _dos_setdrive(old_drv, &dummy);
       if(tpr == NULL) {
          beeper(3); goto select_dir;
       }
       if(dir_chdir(__homes)) {
          beeper(3);
          tpr = dir_getcwd(__homes, 256);
       }
    }
    #endif // #ifdef _IBM_PCDOS_
    else {
       if(   ((k==1 && onlydir) || (k == 0 && !onlydir))
          && __flst[k].name[0] == '.') {
              dirstrip(__homes);
       }
       else {
          i = strlen(__homes);
          if(__homes[i-1] != DIRSEPR) {
             __homes[i] = DIRSEPR;
             __homes[i+1] = 0;
          }
          strcat(__homes, __flst[k].name);
          i = strlen(__homes);
          __homes[i-1] = 0;
       }
       for(i=__files-1; i>=tlp; i--) free(__prs[i]);
    }
  } while( k >= 0);

  dir_chdir(local_dir);
  #ifdef _IBM_PCDOS_
//  _dos_setdrive(old_drv, &dummy);
  #endif
  return 0;
}


