/****************************************************************
*   Utility routines for Generic C                              *
*****************************************************************
*   Text Based Menu and Windowing  (header file)                *
*                                                               *
*                                     Yudong Yang               *
*     Dept. of Computer Science and Technology, TsingHua Univ.  *
*   File : menu.h                                               *
****************************************************************/

/* maximum nest level of menu */
#ifndef _MENU_INC
#define _SCREEN_VAR
#include "screen.h"
#undef  _SCREEN_VAR
#define _MENU_INC

#ifndef  NULL
  #define  NULL  ((void *)0)
#endif
#define     maxnestlevel    20
#define     maxmenustatus   256

/*  text color defination */
#define     BACKGROUND      0x07
#define     SHADOW          0x08
#define     CHOOSEDTITLE    0x07
#define     BUTTONDOWN      0x2f
#define     BUTTONUP        0x20
#define     MENUBACK        0x70
#define     MENUGREY        0x78
#define     CHOOSEDGREY     0x08
#define     QUICKKEY        0x7f
#define     QUICKKEYCHOOSED 0x0f
#define     MENUEDGE        0x7e
#define     CHOISECOMMENT   0x70
#define     DIALOGEDGE      0x3b
#define     DIALOGBACK      0x30
#define     DIALOGTITLE     0x3a
#define     DIALOGTEXT      0x30
#define     ERRORMESS       0x34
#define     LISTEDGE        0x3b
#define     LISTTEXT        0x30
#define     LISTSEL         0x2f
#define     BUSYMESS        0x9f
#define     MEMMESS         0x9e


#define           ESCAPE_CHAR       '&'

// common broadcast message
#define           MENUINIT          100
#define           MENUCLOSE         101
#define           MENUREPAINT       102

/* menu attribute defination, see menutype */
#define     mvertical       1     /* vertical menu */
#define     mfullwidth      2     /* full screen width horizontal menu */
#define     mshadow         4     /* manu with shadow */
#define     mframe          8     /* menu width frame */
#define     mselect        16     /* menu which can only select one (marked)*/


/* menufunction return value */
/* Can be ORed to represent more than one value */
#define         refresh_screen          1  // refresh display screen if possible
#define         close_caller_menu       2  // close present pull down menu
#define         menu_repaint            4  // 



//#ifndef mword
//typedef unsigned short int    mword;
//typedef unsigned char         mbyte;
//#endif

/* menu choise type  */
typedef struct menustruct menutype; 

typedef struct {
  char * string;        /* menu string */
  int   message;        /* choise message */
  menutype * menu;      /* sub menu, set to NULL if no submenu*/
  char * comment;       /* comment for this choise */
  int qkindex;          /* quick key index in string */
 } choisetype ;

struct menustruct{
     mword menutype;        /* menu attribute      */
     choisetype * menulist;
     int choise;
     int barlen;
     int items;
     int opt;              // optional used, as in case of selectors
 };

// menu is defined as
//  choisetype mychoise[] = { 
//                 {"Menu choise &string 1", message_value1,  
//                    NULL or (void *)&submenu, "Comment here "}, // quick index is auto filled
//                 ...  next choise,
//                 NULL };
//  menutype mymenu={combined_attributes_here, &mychoise, 0, 0, 0};


#define MNULL   ((void *)0)


#define menu_grey     1
#define menu_checked  2


// initlize menu system, frame_selection defines which frame char set is used.
int menuinit(int frame_selection);

// exit menu system. You should call this when exit program
int menuexit(void);

// do that menu at (starty, startx), message callback is callbackfunc
int domenu(menutype * menu, mword starty, mword startx , 
           int callbackfunc(int message));

// raise a message to menu system
int raise_message(int message);


// show busy status  busy == 0 is not busy,
// busy =1-100 is percent finished, busy>100 is busy always
void menu_busy(short busy);

// set status of the menu item defined as message 
int set_menu_status(int message, int stat);

// clear status of the menu item defined as message 
int reset_menu_status(int message, int stat);

// get status of the menu item defined as message 
int get_menu_status(int message);

//Show message at the center of row and shadow is on row+1,  attribute is attr
void putnotice(mword row, char * message, mword att);

// say a dialog box return the button index if choosed
// message string start by '@' is a button, '*' is left aligned message
// others are middle placed string.
int dialogbox(char * message[]);

/*#define BTN_OK      0
#define BTN_CANCEL  1
int messagebox(char * message, int flag);
*/

// input a string that has max length 'lth', 
// prompt user by 'comment', if 'keeps', do not clear contents of 'str' at first
int get_string(char *str, int lth, char *comment, int keeps);

// get file name into 'fname', append 'ext' if needed, 
// prompt user by 'comment'
int getfilename(char * fname, char * ext, char *comment);

// input an integer that has max length 'lth', 
// prompt user by 'comment', if 'keeps', do not clear contents of 'str' at first
int getnumber(char *comment);

// input a float point number that has max length 'lth', 
// prompt user by 'comment', if 'keeps', do not clear contents of 'str' at first
double getfloat(char *comment);

// show a list in a 'title'd window and do choosing.
// when choosing bar moved or choosed, 
//  call callbackfunc(index_of_item_start_from_1),
// if callbackfunc  return 0 or ESC key pressed then exit 
// return: normally is current selection index_of_item_start_from_1  
//         or is -1 if ESC pressed
int selectlist(char **list, int starty, int startx , int ysz, 
               int xsz, int callbackfunc(int message), char *title);

// select dir/file with file mask, 
// if onlydir != 0 then only choose directories
int dirsel(char *masks, char *res, char *title, int onlydir);

#endif

