/****************************************************************
*   Utility routines for Generic C                              *
*****************************************************************
*   Text Based Screen Driver  (header file)                     *
*                                                               *
*                                     Yudong Yang               *
*     Dept. of Computer Science and Technology, TsingHua Univ.  *
*   File : screen.h                                             *
****************************************************************/

#ifndef     _SCREEN_INC
#define     _SCREEN_INC
/* cursor controls */
#define     CURSOR_OVERLAY  -1
#define     CURSOR_INSERT   -2
#define     CURSOR_OFF       0
#define     CURSOR_ON        1
#define     CURSOR_RESET     128

/* key code defination */
//$$$ these key codes are system dependent
#define     ESC     0x011b
#define     UP      0x4800
#define     DOWN    0x5000
#define     LEFT    0x4b00
#define     RIGHT   0x4d00
#define     ENTER   0x1c0d
#define     TAB     0x0f09
#define     BACKSPACE   0x0e08
#define     SPACEBAR    0x3920
#define     DELETE  0x5300
#define     INSERT  0x5200
#define     F1      0x3b00
#define     F2      0x3c00
#define     F3      0x3d00
#define     F4      0x3e00
#define     F5      0x3f00
#define     F6      0x4000
#define     F7      0x4100
#define     F8      0x4200
#define     F9      0x4300
#define     F10     0x4400
#define     F11     0x8500
#define     F12     0x8600
#define     SFT_F1  0x5400
#define     SFT_F2  0x5500
#define     SFT_F3  0x5600
#define     SFT_F4  0x5700
#define     SFT_F5  0x5800
#define     SFT_F6  0x5900
#define     SFT_F7  0x5a00
#define     SFT_F8  0x5b00
#define     SFT_F9  0x5c00
#define     SFT_F10 0x5d00
#define     SFT_F11 0x8700
#define     SFT_F12 0x8800

#define     CTL_F1  0x5e00
#define     CTL_F2  0x5f00
#define     CTL_F3  0x6000
#define     CTL_F4  0x6100
#define     CTL_F5  0x6200
#define     CTL_F6  0x6300
#define     CTL_F7  0x6400
#define     CTL_F8  0x6500
#define     CTL_F9  0x6600
#define     CTL_F10 0x6700
#define     CTL_F11 0x8900
#define     CTL_F12 0x8a00

#define     ALT_F1  0x6800
#define     ALT_F2  0x6900
#define     ALT_F3  0x6a00
#define     ALT_F4  0x6b00
#define     ALT_F5  0x6c00
#define     ALT_F6  0x6d00
#define     ALT_F7  0x6e00
#define     ALT_F8  0x6f00
#define     ALT_F9  0x7000
#define     ALT_F10 0x7100
#define     ALT_F11 0x8b00
#define     ALT_F12 0x8c00

#define     HOME    0x4700
#define     END_KEY 0x4f00
#define     PGUP    0x4900
#define     PGDN    0x5100
#define     CTRLLEFT    0x7300
#define     CTRLRIGHT   0x7400
#define     CTL_BREAK   0xffff          /* really is 0x0000 */


//  test_shift  flags
#define           SHIFT_SYSREQ_DN   0x8000
#define           SHIFT_CAP_DN      0x4000
#define           SHIFT_NUM_DN      0x2000
#define           SHIFT_SLK_DN      0x1000
#define           SHIFT_ALT_R       0x0800
#define           SHIFT_CTRL_R      0x0400
#define           SHIFT_ALT_L       0x0200
#define           SHIFT_CTRL_L      0x0100

#define           SHIFT_INS_ON      0x0080
#define           SHIFT_CAP_ON      0x0040
#define           SHIFT_NUM_ON      0x0020
#define           SHIFT_SLK_ON      0x0010
#define           SHIFT_ALT         0x0008
#define           SHIFT_CTRL        0x0004
#define           SHIFT_LEFT        0x0002
#define           SHIFT_RIGHT       0x0001

/* scroll bar direction define */
#define         vertical_scroll_bar     1
#define         horizontal_scroll_bar   2

#ifndef mword
typedef unsigned short int    mword;
typedef unsigned char         mbyte;
typedef unsigned long         mdword;
#endif

/* macro function defination */
#define     trunccursor(row, col)      \
   if(row >= scrymax) row = scrymax-1; \
   if(col >= scrxmax) col = scrxmax-1; 
/* end of macro defination */   


int open_screen(int framesel);
void reset_screen(void);
void close_screen(void);
void screen_controlbreak(int req);        // control break process
int mygetch(void);
int test_shift(void);
int testkey(void);


void beeper(int i);
void crt_cls(mword ulr, mword ulc, mword brr, mword brc, mword att);
void crt_grc(mword * row, mword * col);
void crt_putch(mword ch , mword attribute );
void crt_puts(char * str, mword attribute );
void crt_putsn(char *str, mword number, mword att);
void crt_src(mword row, mword column );
int cursor(int mode);
void clr_eol(mword  att);
mword crt_getch(mword row, mword col);
void crt_getblock(mword * buffer, mword ulr, mword ulc, mword brr, mword brc);
void crt_putblock(mword * buffer, mword ulr, mword ulc, mword brr, mword brc);
void crt_scroll_up(mword ulr, mword ulc, mword brr, mword brc, mword att, int n);
void crt_scroll_down(mword ulr, mword ulc, mword brr, mword brc, mword att, int n);
void crt_scroll(mword ulr, mword ulc, mword brr, mword brc, mword att, int n);

int mygets(char * line, mword att);
int myeditline(char * line, mword att, int maxlth, int winlen);
int drawscrollbar(mword srow, mword scol, mword lth, mword dirc, mword attrib);
int setscrollbar(mword srow, mword scol, mword lth, mword dirc, mword attrib, mword oldpos, mword newpos);

int shadow(mword ulr, mword ulc, mword brr, mword brc, mword att);
int frame(mword ulr, mword ulc, mword brr, mword brc, mword att) ;
// open a window to display your messages return handle of a window, -1 if fail
int open_window(char *name, mword ulr, mword ulc, mword brr, mword brc, mword edgeattr, mword cntattr);
// close that window
int close_window(int wnum);
// print something in window 'wnum' with attribute 'attr'
int win_printf(int wnum, mword attr, const char *fmt, ...);
// putchar in window 'wnum' with attribute 'attr'
int win_putchar(int wnum, mword attr, const char ch);
// repaint the contents of window wnum
void repaint_window(int wnum);

#ifdef _SCREEN_VAR
extern mword scrxmax, scrymax;             /* screen size */
extern mword controlbreakflag;
extern char  _checker, _selector;  // Checker and Marker of current frame type
#endif

#endif
