/****************************************************************
*   Utility routines for C                                      *
*****************************************************************
*   A simple syntax parser                                      *
*                                                               *
*                                     Yudong Yang               *
*     Dept. of Computer Science and Technology, TsingHua Univ.  *
*   File : syntaxps.c                                           *
****************************************************************/
/* 
 * File creation date: 1996.11.19
 * Modification: 
 */

#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "utility.h"

#define _SYN_NO_EXT
#include "syntaxps.h"

#define MAX_SYN_QUE 100

// a simple syntax tag queue 

syntax_queue * _synq;
int _sqlen=0;

// scan the syntax of a line, strip comment and return the number of items
int scan_syntax(char *bf, char comment_char)
{ 
  int i,j,k, p, ft;
  int conti;
  char *t;
  enum {start=0, msign, maindigi, dot, rmdigi, expstart, expsign, expdigi} numst;

  _sqlen=0;
  if(!_synq) _synq = malloc(MAX_SYN_QUE*sizeof(syntax_queue));
  if(!_synq) return -1;
  k=strstrip(bf);
  if(k<=0) return 0;
  conti=1; i=0; p=0; 
  while(conti) {
    if(isspace(bf[i]) || bf[i] == comment_char || i == k) {
       if(p<i) {
         _synq[_sqlen].s_type = s_string;  // first we assume 
         _synq[_sqlen].ptr = bf+p;
         _synq[_sqlen].lth = i-p;
         _sqlen ++;
       }
       if(i >= k) conti = 0;
       else {
          while(isspace(bf[i]) && ++i<k); // jump spaces
          if(bf[i] == comment_char) { bf[i]=0; i=strstrip(bf); conti=0; }
          p = i;
       }
    }
    else i++;
  }
  for(j=0; j<_sqlen; j++) {
    k = _synq[j].lth;
    t = _synq[j].ptr;
    if(isalpha(t[0]) || t[0]=='_') {
      for(i=0; i<k && (isalnum(t[i]) || t[i]=='_'); i++);
      if(i == k) _synq[j].s_type = s_ident; // passed identifier check
      continue; // otherwise it can not be a number
    }
    numst = start; p=1; ft=0; // p is pass. ft is float number
    for(i=0; i<k && p; i++) {
      switch(numst) {
        case start:
          if(t[i] == '+' || t[i] == '-') {numst = msign;}
          else if(t[i] >= '0' && t[i] <= '9') {numst = maindigi;}
          else if(t[i] == '.') {numst = rmdigi; ft = 1;}
          else p = 0; 
          break;
        case msign:  
          if(t[i] >= '0' && t[i] <= '9') {numst = maindigi;}
          else if(t[i] == '.') {numst = dot; ft = 1;}
          else p = 0;
          break;
        case maindigi:  
          if(t[i] >= '0' && t[i] <= '9') {numst = maindigi;}
          else if(t[i] == '.') {numst = rmdigi; ft = 1;}
          else if(t[i] == 'e' || t[i] == 'E') {numst = expstart; ft = 1;}
          else p = 0;
          break;
        case dot:
        case rmdigi:
          if(t[i] >= '0' && t[i] <= '9') {numst = rmdigi; ft = 1;}
          else if(t[i] == 'e' || t[i] == 'E') {numst = expstart; ft = 1;}
          else p = 0;
          break;
        case expstart:
          if(t[i] == '+' || t[i] == '-') {numst = expsign;}
          else if (t[i] >= '0' && t[i] <= '9') {numst = expdigi;}
          else p = 0; 
          break;
        case expsign:
        case expdigi:
          if(t[i] >= '0' && t[i] <= '9') {numst = expdigi;}
          else p = 0; 
        default:
          p = 0;
      }
    }
    if(p && (numst == maindigi || numst == rmdigi || numst == expdigi)) {
       if(ft)  _synq[j].s_type = s_float;
       else  _synq[j].s_type = s_int;
    }
  }  
  return _sqlen;
}

